#include <stdio.h>
#include <string.h>
#include "DKC.h"
#include "MEM.h"
#include "DBG.h"
#include "dvm_pri.h"

static void
implement_diksam_function(DVM_VirtualMachine *dvm, int dest_idx,
                          ExecutableEntry *ee, int src_idx)
{
    dvm->function[dest_idx]->u.diksam_f.executable
        = ee;
    dvm->function[dest_idx]->u.diksam_f.index = src_idx;
}

static void
add_functions(DVM_VirtualMachine *dvm, ExecutableEntry *ee)
{
    int src_idx;
    int dest_idx;
    int add_func_count = 0;
    DVM_Boolean *new_func_flags;

    new_func_flags = MEM_malloc(sizeof(DVM_Boolean)
                                * ee->executable->function_count);

    for (src_idx = 0; src_idx < ee->executable->function_count; src_idx++) {
        for (dest_idx = 0; dest_idx < dvm->function_count; dest_idx++) {
            if (!strcmp(dvm->function[dest_idx]->name,
                        ee->executable->function[src_idx].name)
                && dvm_compare_package_name(dvm->function[dest_idx]
                                            ->package_name,
                                            ee->executable->function
                                            [src_idx].package_name)) {
                if (ee->executable->function[src_idx].is_implemented
                    && dvm->function[dest_idx]->is_implemented) {
                    char *package_name;

                    if (dvm->function[dest_idx]->package_name) {
                        package_name = dvm->function[dest_idx]->package_name;
                    } else {
                        package_name = "";
                    }
                    dvm_error(NULL, NULL, NO_LINE_NUMBER_PC,
                              FUNCTION_MULTIPLE_DEFINE_ERR,
                              STRING_MESSAGE_ARGUMENT, "package",
                              package_name,
                              STRING_MESSAGE_ARGUMENT, "name",
                              dvm->function[dest_idx]->name,
                              MESSAGE_ARGUMENT_END);
                }
                new_func_flags[src_idx] = DVM_FALSE;
                if (ee->executable->function[src_idx].is_implemented) {
                    implement_diksam_function(dvm, dest_idx, ee, src_idx);
                }
                break;
            }
        }
        if (dest_idx == dvm->function_count) {
            new_func_flags[src_idx] = DVM_TRUE;
            add_func_count++;
        }
    }
    dvm->function
        = MEM_realloc(dvm->function,
                      sizeof(Function*)
                      * (dvm->function_count + add_func_count));

    for (src_idx = 0, dest_idx = dvm->function_count;
         src_idx < ee->executable->function_count; src_idx++) {
        if (!new_func_flags[src_idx])
            continue;

        dvm->function[dest_idx] = MEM_malloc(sizeof(Function));
        if (ee->executable->function[src_idx].package_name) {
            dvm->function[dest_idx]->package_name
                = MEM_strdup(ee->executable->function[src_idx].package_name);
        } else {
            dvm->function[dest_idx]->package_name = NULL;
        }
        dvm->function[dest_idx]->name
            = MEM_strdup(ee->executable->function[src_idx].name);
        dvm->function[dest_idx]->kind = DIKSAM_FUNCTION;
        dvm->function[dest_idx]->is_implemented
            = ee->executable->function[src_idx].is_implemented;
        if (ee->executable->function[src_idx].is_implemented) {
            implement_diksam_function(dvm, dest_idx, ee, src_idx);
        }
        dest_idx++;
    }
    dvm->function_count += add_func_count;
    MEM_free(new_func_flags);
}

static int
search_function(DVM_VirtualMachine *dvm, char *package_name, char *name)
{
    int i;

    for (i = 0; i < dvm->function_count; i++) {
        if (dvm_compare_package_name(dvm->function[i]->package_name,
                                     package_name)
            && !strcmp(dvm->function[i]->name, name)) {
            return i;
        }
    }
    return FUNCTION_NOT_FOUND;
}

static int
search_class(DVM_VirtualMachine *dvm, char *package_name, char *name)
{
    int i;

    for (i = 0; i < dvm->class_count; i++) {
        if (dvm_compare_package_name(dvm->class[i]->package_name,
                                     package_name)
            && !strcmp(dvm->class[i]->name, name)) {
            return i;
        }
    }
    dvm_error(NULL, NULL, NO_LINE_NUMBER_PC, CLASS_NOT_FOUND_ERR,
              STRING_MESSAGE_ARGUMENT, "name", name,
              MESSAGE_ARGUMENT_END);
    return 0; /* make compiler happy */
}

static void
convert_code(DVM_VirtualMachine *dvm, DVM_Executable *exe,
             DVM_Byte *code, int code_size, DVM_Function *func)
{
    int i;
    int j;
    OpcodeInfo *info;
    int src_idx;
    unsigned int dest_idx;

    for (i = 0; i < code_size; i++) {
        if (code[i] == DVM_PUSH_STACK_INT
            || code[i] == DVM_PUSH_STACK_DOUBLE
            || code[i] == DVM_PUSH_STACK_OBJECT
            || code[i] == DVM_POP_STACK_INT
            || code[i] == DVM_POP_STACK_DOUBLE
            || code[i] == DVM_POP_STACK_OBJECT) {
            int parameter_count;

            DBG_assert(func != NULL, ("func == NULL!\n"));
            if (func->is_method) {
                parameter_count = func->parameter_count + 1; /* for this */
            } else{
                parameter_count = func->parameter_count;
            }

            src_idx = GET_2BYTE_INT(&code[i+1]);
            if (src_idx >= parameter_count) {
                dest_idx = src_idx + CALL_INFO_ALIGN_SIZE;
            } else {
                dest_idx = src_idx;
            }
            SET_2BYTE_INT(&code[i+1], dest_idx);

        } else if (code[i] == DVM_PUSH_FUNCTION) {
            int idx_in_exe;
            unsigned int func_idx;

            idx_in_exe = GET_2BYTE_INT(&code[i+1]);
            func_idx = search_function(dvm,
                                       exe->function[idx_in_exe].package_name,
                                       exe->function[idx_in_exe].name);
            SET_2BYTE_INT(&code[i+1], func_idx);

        } else if (code[i] == DVM_NEW) {
            int idx_in_exe;
            unsigned int class_idx;

            idx_in_exe = GET_2BYTE_INT(&code[i+1]);
            class_idx = search_class(dvm,
                                     exe->class_definition[idx_in_exe]
                                     .package_name,
                                     exe->class_definition[idx_in_exe].name);
            SET_2BYTE_INT(&code[i+1], class_idx);
        }
        info = &dvm_opcode_info[code[i]];
        for (j = 0; info->parameter[j] != '\0'; j++) {
            switch (info->parameter[j]) {
            case 'b':
                i++;
                break;
            case 's': /* FALLTHRU */
            case 'p':
                i += 2;
                break;
            default:
                DBG_assert(0, ("param..%s, j..%d", info->parameter, j));
            }
        }
    }
}

static void
add_static_variables(ExecutableEntry *entry, DVM_Executable *exe)
{
    int i;

    entry->static_v.variable
        = MEM_malloc(sizeof(DVM_Value) * exe->global_variable_count);
    entry->static_v.variable_count = exe->global_variable_count;

    for (i = 0; i < exe->global_variable_count; i++) {
        if (exe->global_variable[i].type->basic_type == DVM_STRING_TYPE) {
            entry->static_v.variable[i].object = NULL_OBJECT_REF;
        }
    }
    for (i = 0; i < exe->global_variable_count; i++) {
        dvm_initialize_value(exe->global_variable[i].type,
                             &entry->static_v.variable[i]);
    }
}

static DVM_Class *
search_class_from_executable(DVM_Executable *exe, DVM_ClassIdentifier *ci)
{
    int i;

    for (i = 0; i < exe->class_count; i++) {
        if (!strcmp(exe->class_definition[i].name, ci->name)) {
            return &exe->class_definition[i];
        }
    }
    DBG_panic(("class %s not found.", ci->name));

    return NULL; /* make compiler happy */
}

static int
set_field_types(DVM_Executable *exe, DVM_Class *pos,
                DVM_TypeSpecifier **field_type, int index)
{
    DVM_Class *next;
    int i;

    if (pos->super_class) {
        next = search_class_from_executable(exe, pos->super_class);
        index = set_field_types(exe, next, field_type, index);
    }
    for (i = 0; i < pos->field_count; i++) {
        field_type[index] = pos->field[i].type;
        index++;
    }

    return index;
}

static void
add_fields(DVM_Executable *exe, DVM_Class *src, ExecClass *dest)
{
    int field_count = 0;
    DVM_Class *pos;

    for (pos = src; ; ) {
        field_count += pos->field_count;
        if (pos->super_class == NULL)
            break;
        pos = search_class_from_executable(exe, src->super_class);
    }
    dest->field_count = field_count;
    dest->field_type = MEM_malloc(sizeof(DVM_TypeSpecifier*) * field_count);
    set_field_types(exe, src, dest->field_type, 0);
}

static void
set_v_table(DVM_VirtualMachine *dvm, DVM_Class *class_p,
            DVM_Method *src, VTableItem *dest, DVM_Boolean set_name)
{
    char *func_name;
    int  func_idx;

    if (set_name) {
        dest->name = MEM_strdup(src->name);
    }
    func_name = dvm_create_method_function_name(class_p->name, src->name);
    func_idx = search_function(dvm, class_p->package_name, func_name);

    if (func_idx == FUNCTION_NOT_FOUND && !src->is_abstract) {
        dvm_error(NULL, NULL, NO_LINE_NUMBER_PC, FUNCTION_NOT_FOUND_ERR,
                  STRING_MESSAGE_ARGUMENT, "name", func_name,
                  MESSAGE_ARGUMENT_END);
    }
    MEM_free(func_name);
    dest->index = func_idx;
}

static int
add_method(DVM_VirtualMachine *dvm, DVM_Executable *exe, DVM_Class *pos,
           DVM_VTable *v_table)
{
    DVM_Class   *next;
    int         i;
    int         j;
    int         super_method_count = 0;
    int         method_count;

    if (pos->super_class) {
        next = search_class_from_executable(exe, pos->super_class);
        super_method_count = add_method(dvm, exe, next, v_table);
    }

    method_count = super_method_count;
    for (i = 0; i < pos->method_count; i++) {
        for (j = 0; j < super_method_count; j++) {
            if (!strcmp(pos->method[i].name, v_table->table[j].name)) {
                set_v_table(dvm, pos, &pos->method[i], &v_table->table[j],
                            DVM_FALSE);
                break;
            }
        }
        /* if pos->method[i].is_override == true,
         * this method implements interface method.
         */
        if (j == super_method_count && !pos->method[i].is_override) {
            v_table->table
                = MEM_realloc(v_table->table,
                              sizeof(VTableItem) * (method_count + 1));
            set_v_table(dvm, pos, &pos->method[i],
                        &v_table->table[method_count], DVM_TRUE);
            method_count++;
            v_table->table_size = method_count;
        }
    }

    return method_count;
}

static DVM_VTable *
alloc_v_table(ExecClass *exec_class)
{
    DVM_VTable *v_table;

    v_table = MEM_malloc(sizeof(DVM_VTable));
    v_table->exec_class = exec_class;
    v_table->table = NULL;

    return v_table;
}

static void
add_methods(DVM_VirtualMachine *dvm, DVM_Executable *exe,
            DVM_Class *src, ExecClass *dest)
{
    int         method_count;
    DVM_VTable  *v_table;
    int         i;
    DVM_Class   *interface;
    int         method_idx;

    v_table = alloc_v_table(dest);
    method_count = add_method(dvm, exe, src, v_table);
    dest->class_table = v_table;
    dest->interface_count = src->interface_count;
    dest->interface_table
        = MEM_malloc(sizeof(DVM_VTable*) * src->interface_count);

    for (i = 0; i < src->interface_count; i++) {
        dest->interface_table[i] = alloc_v_table(dest);
        interface = search_class_from_executable(exe, &src->interface[i]);
        dest->interface_table[i]->table
            = MEM_malloc(sizeof(VTableItem) * interface->method_count);
        dest->interface_table[i]->table_size = interface->method_count;
        for (method_idx = 0; method_idx < interface->method_count;
             method_idx++) {
            set_v_table(dvm, src, &interface->method[method_idx],
                        &dest->interface_table[i]->table[method_idx],
                        DVM_TRUE);
        }
    }
}

static void
add_class(DVM_VirtualMachine *dvm, DVM_Executable *exe,
          DVM_Class *src, ExecClass *dest)
{
    add_fields(exe, src, dest);
    add_methods(dvm, exe, src, dest);
}

static void
add_classes(DVM_VirtualMachine *dvm, DVM_Executable *exe)
{
    int src_idx;
    int dest_idx;
    int add_class_count = 0;
    DVM_Boolean *new_class_flags;

    new_class_flags = MEM_malloc(sizeof(DVM_Boolean)
                                 * exe->class_count);

    for (src_idx = 0; src_idx < exe->class_count; src_idx++) {
        for (dest_idx = 0; dest_idx < dvm->class_count; dest_idx++) {
            if (!strcmp(dvm->class[dest_idx]->name,
                        exe->class_definition[src_idx].name)
                && dvm_compare_package_name(dvm->class[dest_idx]
                                            ->package_name,
                                            exe->class_definition[src_idx]
                                            .package_name)) {
                if (exe->class_definition[src_idx].is_implemented
                    && dvm->class[dest_idx]->is_implemented) {
                    char *package_name;

                    if (dvm->class[dest_idx]->package_name) {
                        package_name = dvm->class[dest_idx]->package_name;
                    } else {
                        package_name = "";
                    }
                    dvm_error(NULL, NULL, NO_LINE_NUMBER_PC,
                              CLASS_MULTIPLE_DEFINE_ERR,
                              STRING_MESSAGE_ARGUMENT, "package",
                              package_name,
                              STRING_MESSAGE_ARGUMENT, "name",
                              dvm->class[dest_idx]->name,
                              MESSAGE_ARGUMENT_END);
                }
                new_class_flags[src_idx] = DVM_FALSE;
                if (exe->class_definition[src_idx].is_implemented) {
                    add_class(dvm, exe, &exe->class_definition[src_idx],
                              dvm->class[dest_idx]);
                }
                break;
            }
        }
        if (dest_idx == dvm->class_count) {
            new_class_flags[src_idx] = DVM_TRUE;
            add_class_count++;
        }
    }
    dvm->class
        = MEM_realloc(dvm->class, sizeof(ExecClass*)
                      * (dvm->class_count + add_class_count));

    for (src_idx = 0, dest_idx = dvm->class_count;
         src_idx < exe->class_count; src_idx++) {
        if (!new_class_flags[src_idx])
            continue;

        dvm->class[dest_idx] = MEM_malloc(sizeof(ExecClass));
        if (exe->class_definition[src_idx].package_name) {
            dvm->class[dest_idx]->package_name
                = MEM_strdup(exe->class_definition[src_idx].package_name);
        } else {
            dvm->class[dest_idx]->package_name = NULL;
        }
        dvm->class[dest_idx]->name
            = MEM_strdup(exe->class_definition[src_idx].name);
        dvm->class[dest_idx]->is_implemented
            = exe->class_definition[src_idx].is_implemented;
        if (exe->class_definition[src_idx].is_implemented) {
            add_class(dvm, exe, &exe->class_definition[src_idx],
                      dvm->class[dest_idx]);
        }
        dest_idx++;
    }
    dvm->class_count += add_class_count;
    MEM_free(new_class_flags);
}

static void
add_executable_to_dvm(DVM_VirtualMachine *dvm, DVM_Executable *executable,
                      DVM_Boolean is_top_level)
{
    int i;
    ExecutableEntry *ee_pos;
    ExecutableEntry *new_entry;
    
    new_entry = MEM_malloc(sizeof(ExecutableEntry));
    new_entry->executable = executable;
    new_entry->next = NULL;

    if (dvm->executable_list == NULL) {
        dvm->executable_list = new_entry;
    } else {
        for (ee_pos = dvm->executable_list; ee_pos->next;
             ee_pos = ee_pos->next)
            ;
        ee_pos->next = new_entry;
    }
    
    add_functions(dvm, new_entry);
    add_classes(dvm, executable);

    convert_code(dvm, executable,
                 executable->code, executable->code_size,
                 NULL);

    for (i = 0; i < executable->function_count; i++) {
        convert_code(dvm, executable,
                     executable->function[i].code,
                     executable->function[i].code_size,
                     &executable->function[i]);
    }

    add_static_variables(new_entry, executable);

    if (is_top_level) {
        dvm->top_level = new_entry;
    }
}

void
DVM_set_executable(DVM_VirtualMachine *dvm, DVM_ExecutablePool *pool)
{
    DVM_ExecutableList *pos;

    dvm->executable_pool = pool;
    for (pos = pool->list; pos; pos = pos->next) {
        add_executable_to_dvm(dvm, pos->executable,
                              (pos->executable == pool->top_level));
    }
}

DVM_VirtualMachine *
DVM_create_virtual_machine(void)
{
    DVM_VirtualMachine *dvm;

    dvm = MEM_malloc(sizeof(DVM_VirtualMachine));
    dvm->stack.alloc_size = STACK_ALLOC_SIZE;
    dvm->stack.stack = MEM_malloc(sizeof(DVM_Value) * STACK_ALLOC_SIZE);
    dvm->stack.pointer_flags
        = MEM_malloc(sizeof(DVM_Boolean) * STACK_ALLOC_SIZE);
    dvm->stack.stack_pointer = 0;
    dvm->heap.current_heap_size = 0;
    dvm->heap.header = NULL;
    dvm->heap.current_threshold = HEAP_THRESHOLD_SIZE;
    dvm->current_executable = NULL;
    dvm->function = NULL;
    dvm->function_count = 0;
    dvm->class = NULL;
    dvm->class_count = 0;
    dvm->executable_pool = NULL;
    dvm->executable_list = NULL;
    dvm->top_level = NULL;

    dvm_add_native_functions(dvm);

    return dvm;
}

void
DVM_add_native_function(DVM_VirtualMachine *dvm,
                        char *package_name, char *func_name,
                        DVM_NativeFunctionProc *proc, int arg_count)
{
    dvm->function
        = MEM_realloc(dvm->function,
                      sizeof(Function*) * (dvm->function_count + 1));

    dvm->function[dvm->function_count] = MEM_malloc(sizeof(Function));
    dvm->function[dvm->function_count]->package_name
        = MEM_strdup(package_name);
    dvm->function[dvm->function_count]->name = MEM_strdup(func_name);
    dvm->function[dvm->function_count]->kind = NATIVE_FUNCTION;
    dvm->function[dvm->function_count]->is_implemented = DVM_TRUE;
    dvm->function[dvm->function_count]->u.native_f.proc = proc;
    dvm->function[dvm->function_count]->u.native_f.arg_count = arg_count;
    dvm->function_count++;
}

static void
make_search_path(char *package_name, char *buf)
{
    int suffix_len;
    int package_len;
    int i;

    suffix_len = strlen(DIKSAM_IMPLEMENTATION_SUFFIX);
    package_len = strlen(package_name);

    DBG_assert(package_len <= FILENAME_MAX - (2 + suffix_len),
               ("package name is too long(%s)", package_name));

    for (i = 0; package_name[i] != '\0'; i++) {
        if (package_name[i] == '.') {
            buf[i] = FILE_SEPARATOR;
        } else {
            buf[i] = package_name[i];
        }
    }
    buf[i] = '\0';
    strcat(buf, DIKSAM_IMPLEMENTATION_SUFFIX);
}

static FILE *
get_dynamic_load_fp(DVM_Executable *exe, Function *caller, int pc,
                    Function *func, char *found_path)
{
    char *search_path;
    char search_file[FILENAME_MAX];
    FILE *fp;
    SearchFileStatus status;
        
    search_path = getenv("DKM_LOAD_SEARCH_PATH");
    if (search_path == NULL) {
        search_path = ".";
    }
    make_search_path(func->package_name, search_file);
    status = dvm_search_file(search_path, search_file, found_path, &fp);
    
    if (status != SEARCH_FILE_SUCCESS) {
        if (status == SEARCH_FILE_NOT_FOUND) {
            dvm_error(exe, caller, pc,
                      LOAD_FILE_NOT_FOUND_ERR,
                      STRING_MESSAGE_ARGUMENT, "file", search_file,
                      MESSAGE_ARGUMENT_END);
        } else {
            dvm_error(exe, caller, pc,
                      LOAD_FILE_ERR,
                      INT_MESSAGE_ARGUMENT, "status", (int)status,
                      MESSAGE_ARGUMENT_END);
        }
    }

    return fp;
}

void
dvm_dynamic_load(DVM_VirtualMachine *dvm,
                 DVM_Executable *caller_exe, Function *caller, int pc,
                 Function *func)
{
    FILE *fp;
    DKC_Compiler *compiler;
    DVM_ExecutableList *pos;
    DVM_ExecutableList *add_top;
    char found_path[FILENAME_MAX];

    fp = get_dynamic_load_fp(caller_exe, caller, pc, func, found_path);
    compiler = DKC_create_compiler();
    add_top = DKC_dynamic_compile(compiler, fp, func->package_name,
                                  dvm->executable_pool,
                                  found_path);

    for (pos = add_top; pos; pos = pos->next) {
        add_executable_to_dvm(dvm, pos->executable, DVM_FALSE);
    }
    DKC_dispose_compiler(compiler);
}
